#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Some general utility functions.

Includes os stuff, dialogs for the main frame,
datetime conversions...
"""

import os
import re
from datetime import datetime
from pathlib import Path
from shutil import copy2
from typing import List, Callable, Sequence, Optional

import cv2
import wx
import wx.adv
from pkg_resources import resource_filename

import story_time

# Paths to app code and temp folder
project_path = Path(os.path.dirname(os.path.realpath(__file__))).parent
icon_path = os.path.join(project_path, "story_time/Icons")
temp_folder = os.path.join(
    project_path, "tmp"
)  #: Temporary folder to store images temporarily.
info_file = resource_filename(__name__, f"Info.txt")

# Path to data (global variables)
# Initialized values should never be used
# Also they should not be imported from the module since they change
# when initializing the app.
data_path = ""  #: The folder containing the image and the xml folder.
img_folder = "if/you/see/this/its/a/bug"  #: The folder where the images are stored.
xml_folder = (
    "if/you/see/this/its/a/bug"  #: The folder where the xml documents are stored.
)

# Set colors
# dark_green = wx.Colour(1, 92, 68)
green = wx.Colour(0, 143, 105)
light_green = wx.Colour(169, 245, 213)
very_light_green = wx.Colour(212, 252, 235)

header_col = light_green
but_bg_col = green
text_bg_col = very_light_green

# Fonts
header_f_info = wx.FontInfo(12).Bold().Family(wx.FONTFAMILY_SWISS)
button_f_info = wx.FontInfo(12).Family(wx.FONTFAMILY_SWISS).Bold()


def update_folder(new_data_path: str) -> None:
    """Update global path variables if data folder is changed.
    """
    global data_path
    global img_folder
    global xml_folder

    data_path = new_data_path
    img_folder = os.path.join(new_data_path, "Img")
    xml_folder = os.path.join(new_data_path, "XML")


def rep_newlines_with_space(string: str) -> str:
    """Removes newlines and replaces them with spaces.

    Also reduces double spaces to single spaces.
    """
    return string.replace("\n", " ").replace("  ", " ")


def create_xml_and_img_folder(base_folder: str) -> None:
    """Create a folder for the XML and the image files in `base_folder`.
    """
    xml_pth = os.path.join(base_folder, "XML")
    if not os.path.isdir(xml_pth):
        os.mkdir(xml_pth)
    img_pth = os.path.join(base_folder, "Img")
    if not os.path.isdir(img_pth):
        os.mkdir(img_pth)
    return


def check_date_time(
    y: int, mon: int, d: int, h: int = 0, minute: int = 0, sec: int = 0
) -> bool:
    """Checks if the provided date and time is valid."""
    try:
        datetime(y, mon, d, h, minute, sec)
    except ValueError:
        return False
    return True


def ask_for_dir(_fun: Callable = None, show: bool = True):
    """Ask the user to select a directory."""
    dial_text = "Choose directory to store Imgs and Text data."
    flags = wx.DD_DEFAULT_STYLE | wx.DD_DIR_MUST_EXIST
    cdDiag = wx.DirDialog(None, dial_text, "", flags)
    if _fun is not None:
        _fun(cdDiag)
    cdDiag.ShowModal() if show else None
    return cdDiag.GetPath()


def get_info_from_file(ask: bool = True, _fun: Callable = None) -> Optional[str]:
    """Read the info file and get necessary information

    Also creates sub-folders if they do not exist.

    Args:
        ask: Whether to ask for a directory.
        _fun: Function to call when dialog pops up. For testing.

    Returns:
        Contents of the file.
    """
    if not os.path.exists(info_file):
        with open(info_file, "w") as f:
            f.write("NoDirectorySpecified")
    with open(info_file) as file:
        data = file.readlines()
        if data != [] and os.path.isdir(data[0]):
            fol_path = data[0]
            create_xml_and_img_folder(fol_path)
            return fol_path
        elif ask:
            files_path = story_time.util.ask_for_dir(_fun)
            create_xml_and_img_folder(files_path)
            return files_path


def write_folder_to_file() -> None:
    """Write the current working directory to file `info_file`.
    """
    with open(info_file, "w") as f:
        f.write(data_path)


def scale_bitmap(bitmap: wx.Bitmap, width: int, height: int) -> wx.Bitmap:
    """Rescales a bitmap to a given size.

    Converting it to image, rescaling and converting it back.
    """
    image = bitmap.ConvertToImage()
    image = image.Scale(width, height, wx.IMAGE_QUALITY_HIGH)
    result = wx.Bitmap(image)
    return result


class PhotoShow(wx.Dialog):
    okay: bool = False
    resized = False

    f_name: str

    def __init__(self, parent, f_name):
        super().__init__(parent, style=wx.DEFAULT_FRAME_STYLE | wx.RESIZE_BORDER)
        self.SetTitle("Large photo view")
        wx.Panel(self)
        self.v_box = wx.BoxSizer(wx.VERTICAL)

        self.f_name = f_name
        b = getImageToShow(f_name)
        self.imageCtrl = wx.StaticBitmap(self, wx.ID_ANY, b)
        self.v_box.Add(self.imageCtrl, 0, wx.ALL | wx.EXPAND, 0)

        self.SetSizer(self.v_box)

        self.SetBackgroundColour(text_bg_col)
        self.Size = wx.Size(500, 500)

        self.Bind(wx.EVT_SIZE, self.OnSize)
        self.Bind(wx.EVT_IDLE, self.OnIdle)

    def OnSize(self, _):
        self.resized = True  # set dirty

    def OnIdle(self, _):
        if self.resized:
            # take action if the dirty flag is set
            s = self.Size
            s_min = s[0] if s[0] < s[1] else s[1]
            self.imageCtrl.SetBitmap(getImageToShow(self.f_name, s_min))
            self.Layout()
            self.resized = False  # reset the flag


class ButtonDialogBase(wx.Dialog):
    """Base class for a dialog with a button."""

    v_box: wx.BoxSizer

    def setup_v_box(self, h_box, h_button, pnl):
        h_box.Add(h_button, flag=wx.LEFT, border=5)
        self.v_box.Add(pnl, proportion=1, flag=wx.ALL | wx.EXPAND, border=5)
        self.v_box.Add(h_box, flag=wx.ALIGN_CENTER | wx.TOP | wx.BOTTOM, border=10)
        self.SetSizer(self.v_box)
        self.SetBackgroundColour(text_bg_col)


class TwoButtonDialogBase(ButtonDialogBase):
    """Base class for a dialog with at least two buttons."""

    def setup(
        self,
        pnl: wx.Panel,
        shoot_label: str,
        cancel_label: str,
        shoot_fun: Callable = None,
        cancel_fun: Callable = None,
    ):
        """Layout setup, to be called in self.InitUI."""
        if shoot_fun is None:
            shoot_fun = self.OnTakePic
        if cancel_fun is None:
            cancel_fun = self.OnClose

        # Buttons
        shootButton = wx.Button(self, label=shoot_label)
        # shootButton.SetBackgroundColour(but_bg_col)
        shootButton.Bind(wx.EVT_BUTTON, shoot_fun)
        cancelButton = wx.Button(self, label=cancel_label)
        cancelButton.Bind(wx.EVT_BUTTON, cancel_fun)
        # cancelButton.SetBackgroundColour(but_bg_col)
        shootButton.SetFont(wx.Font(button_f_info))
        cancelButton.SetFont(wx.Font(button_f_info))

        # Layout
        h_box = wx.BoxSizer(wx.HORIZONTAL)
        h_box.Add(shootButton)

        self.setup_v_box(h_box, cancelButton, pnl)

    def OnTakePic(self, e):
        pass

    def OnClose(self, e):
        pass


class CustomMessageDialog(TwoButtonDialogBase):
    okay: bool = False

    def __init__(
        self,
        message: str = None,
        title: str = "Message",
        *args,
        cancel_only: bool = False,
        ok_label: str = "Ok",
        cancel_label: str = "Cancel",
        **kw,
    ):
        super().__init__(*args, **kw)
        self.SetTitle(title)
        pnl = wx.Panel(self)
        self.v_box = wx.BoxSizer(wx.VERTICAL)

        if message is not None:
            stMsg = wx.StaticText(self, -1, message)
            # stMsg.SetBackgroundColour(green)
            self.v_box.Add(stMsg, 1, wx.ALIGN_CENTER | wx.ALL, 10)

        self.SetSizer(self.v_box)

        if cancel_only:
            h_box = wx.BoxSizer(wx.HORIZONTAL)
            cancel_butt = wx.Button(self, label=cancel_label)
            cancel_butt.Bind(wx.EVT_BUTTON, self.OnOK)
            cancel_butt.SetFont(wx.Font(button_f_info))
            self.setup_v_box(h_box, cancel_butt, pnl)
        else:
            self.setup(pnl, ok_label, cancel_label, self.OnOK, self.OnClose)

    def OnOK(self, _):
        print("Okay")
        self.okay = True
        self.Close()

    def OnClose(self, _):
        print("cancel")
        self.Close()


class ChangeDateDialog(TwoButtonDialogBase):
    """Date and Time picker dialog"""

    dt: wx.DateTime
    cal: wx.adv.CalendarCtrl
    timePicker: wx.adv.TimePickerCtrl

    def __init__(self, *args, **kw):
        super(ChangeDateDialog, self).__init__(*args, **kw)

        self.dt = wx.DateTime.Now()
        self.InitUI()
        self.SetSize((400, 300))
        self.SetTitle("Change Date of entry")

    def InitUI(self) -> None:
        pnl = wx.Panel(self)

        # Calendar and Time Picker
        sb = wx.StaticBox(pnl, label="Select Date and Time")
        sbs = wx.StaticBoxSizer(sb, orient=wx.HORIZONTAL)
        self.cal = wx.adv.CalendarCtrl(pnl)
        sbs.Add(self.cal, flag=wx.ALL, border=5)
        self.timePicker = wx.adv.TimePickerCtrl(pnl)
        sbs.Add(self.timePicker, flag=wx.ALL, border=5)
        pnl.SetSizer(sbs)

        # Add buttons
        self.v_box = wx.BoxSizer(wx.VERTICAL)
        self.setup(pnl, "Ok", "Cancel", self.OnOK, self.OnClose)

    def OnClose(self, e) -> None:
        print("Closed Change Date Dialog")
        self.Close()

    def OnOK(self, _) -> None:
        """Sets `self.dt` to the current time."""
        timeTuple = self.timePicker.GetTime()
        self.dt = self.cal.GetDate()
        self.dt.SetHour(timeTuple[0])
        self.dt.SetMinute(timeTuple[1])
        self.dt.SetSecond(timeTuple[2])
        print(self.dt)
        self.Close()


class PhotoWithSameDateExistsDialog(ButtonDialogBase):
    """Dialog that pops up if you want to save an image, but
    there exists an image with the same associated time."""

    img: wx.StaticBitmap
    bmp_shown: wx.Bitmap

    next_button: wx.Button
    prev_button: wx.Button

    def __init__(
        self,
        file_list: Sequence,
        parent=None,
        title: str = "Image with same date already exists",
    ):
        super(PhotoWithSameDateExistsDialog, self).__init__(parent, title=title)

        self.file_list = file_list
        self.chosenImgInd = None
        self.shownImgInd = 0
        self.n_files = len(file_list)
        self.InitUI()
        self.SetSize((400, 300))
        self.SetTitle(title)

    def InitUI(self) -> None:

        pnl = wx.Panel(self)
        self.v_box = wx.BoxSizer(wx.VERTICAL)

        # Current Image
        self.bmp_shown = self.get_img_at_ind(0)
        self.img = wx.StaticBitmap(self, -1, self.bmp_shown)
        self.v_box.Add(self.img, proportion=0, flag=wx.ALL, border=5)
        self.img.Show()

        # Buttons
        h_box = wx.BoxSizer(wx.HORIZONTAL)
        selectButton = wx.Button(self, label="Select")
        newButton = wx.Button(self, label="New")

        h_box.Add(selectButton)

        if self.n_files > 1:  # No next / previous if there is only one
            self.next_button = wx.Button(self, label="Next")
            self.prev_button = wx.Button(self, label="Previous")
            self.prev_button.Disable()
            h_box.Add(self.prev_button)
            h_box.Add(self.next_button)
            self.next_button.Bind(wx.EVT_BUTTON, self.OnNext)
            self.prev_button.Bind(wx.EVT_BUTTON, self.OnPrev)

        self.setup_v_box(h_box, newButton, pnl)
        selectButton.Bind(wx.EVT_BUTTON, self.OnSelect)

        newButton.Bind(wx.EVT_BUTTON, self.OnNew)

    def get_img_at_ind(self, ind):
        pth = os.path.join(img_folder, self.file_list[ind])
        return getImageToShow(pth, size=100, border=5)

    def updateImg(self) -> None:
        ind = self.shownImgInd
        img_to_show = self.get_img_at_ind(ind)
        self.img.SetBitmap(img_to_show)

    def OnNext(self, _) -> None:
        print("Next Image")
        self.shownImgInd += 1
        if self.shownImgInd == self.n_files - 1:
            self.next_button.Disable()
        if self.shownImgInd == 1:
            self.prev_button.Enable()
        self.updateImg()
        print("next img")

    def OnPrev(self, _) -> None:
        self.shownImgInd -= 1
        if self.shownImgInd == 0:
            self.prev_button.Disable()
        if self.shownImgInd == self.n_files - 2:
            self.next_button.Enable()
        self.updateImg()
        print("Prev img")

    def OnSelect(self, _):
        self.chosenImgInd = self.shownImgInd
        self.Close()

    def OnClose(self, _):
        print("Closed Change Date Dialog")
        self.Close()

    def OnNew(self, _):
        self.chosenImgInd = -1  # This means new
        self.Close()


def datetime_to_wx_datetime(date: datetime) -> wx.DateTime:
    """Converts a python datetime to a wx.DateTime.
    """
    year = date.year
    month = date.month
    day = date.day
    hour = date.hour
    return wx.DateTime(day, month - 1, year, hour, date.minute, date.second)


def get_file_modified_wx_dt(curr_file: str) -> wx.DateTime:
    """Read the date modified time of a file, convert it to wx.DateTime
    and return it.
    """
    modify_time = os.path.getmtime(curr_file)
    return datetime_to_wx_datetime(datetime.fromtimestamp(modify_time))


def pad_int_str(int_to_pad: int, n: int = 2) -> str:
    """Converts an integer to a string, padding with leading zeros
    to get n characters. Intended for date and time formatting.
    """
    base = str(int_to_pad)
    for k in range(n - 1):
        if int_to_pad < 10 ** (k + 1):
            base = "0" + base
    return base


def get_img_name_from_time(wx_dt: wx.DateTime) -> str:
    """Gives the image basename from the wx.DateTime.
    """
    name = (
        "IMG_"
        + pad_int_str(wx_dt.GetYear(), 4)
        + pad_int_str(wx_dt.GetMonth() + 1)
        + pad_int_str(wx_dt.GetDay())
    )
    name = (
        f"{name}_"
        + pad_int_str(wx_dt.GetHour())
        + pad_int_str(wx_dt.GetMinute())
        + pad_int_str(wx_dt.GetSecond())
    )
    return name


def extract_date_from_image_name(img_file: str) -> Optional[wx.DateTime]:
    """Extracts the date information from the image name.

    Tries to extract the date and optionally the time.
    If extracted date is not valid or there is no date
    contained in the name, returns None.
    """
    p, filename = os.path.split(img_file)
    # Extract all numbers from string
    nums = re.findall(r"\d+", filename)
    num_nums = len(nums)
    if num_nums < 1:
        print("No Date extracted!")
        return None
    date = int(nums[0])
    # Year must be greater than 1
    if date < 10000:
        print("Invalid date.")
        return None
    year = date // 10000
    rem = date - year * 10000
    month = rem // 100
    day = rem - month * 100
    hour = mins = sec = 0

    # Extract time
    if len(nums) > 1:
        num_curr = int(nums[1])
        len_num = len(nums[1])
        if len_num == 6:
            sec = num_curr % 100
            num_curr = num_curr // 100
        if len_num == 4 or len_num == 6:
            mins = num_curr % 100
            hour = num_curr // 100

    if not check_date_time(year, month, day, hour, mins, sec):
        print("Invalid date extracted!")
        return None

    # Check if date is valid
    wx_dt = wx.DateTime(day, month - 1, year, hour, mins, sec)
    return wx_dt if wx_dt.IsValid() else None


def get_time_from_file(img_file: str) -> wx.DateTime:
    """Tries to extract the date from a filename.

    If there is none, returns the modified time
    of the file as wx.DateTime.
    """
    name_date = extract_date_from_image_name(img_file)
    if name_date is None:
        name_date = get_file_modified_wx_dt(img_file)
    return name_date


def find_all_imgs_with_same_date(imgs_folder: str, img_name: str) -> List[str]:
    """Finds all files starting with string "img_name"
    in folder "img_folder".
    """
    res = []
    for f in os.listdir(imgs_folder):
        if f.startswith(img_name):
            res.append(f)
    return res


def find_new_name(
    img_name: str, same_date_file_list: List[str], ext: str, max_n_imgs: int = 10000
) -> str:
    """Chooses a new filename that is not in the list and
    contains the img_name in the beginning by appending an
    int to the filename separated by an underscore.
    """
    for k in range(max_n_imgs):
        new_name = img_name + "_" + str(k) + ext
        try:
            same_date_file_list.index(new_name)
        except ValueError:
            return img_name + "_" + str(k)

    # If there are already more than `max_n_imgs` images, gives up.
    raise ValueError("ERROR: Way too many fucking images.")


def copy_img_file_to_imgs(lf: str, img_date=None, photo_diag_fun: Callable = None):
    """Copy an image to the Img folder.

    If there exists one with
    the same date and time, a dialog pops up
    to let the user select if he wants to add
    text to an existing image or save the image
    If he doesn't decide, returns None
    """
    # Get date of image and find all images with same date
    imgDate = get_time_from_file(lf) if img_date is None else img_date
    imgName = get_img_name_from_time(imgDate)
    sameDateFileList = find_all_imgs_with_same_date(img_folder, imgName)
    print(sameDateFileList)

    # Get image type
    _, file_extension = os.path.splitext(lf)

    # Check if file with same datetime already exists
    if len(sameDateFileList) > 0:
        # Let the user look at the images and decide if he wants
        # to add the text to an already existing one instead of
        # adding the image in preview to the images.
        phDiag = PhotoWithSameDateExistsDialog(sameDateFileList)
        if photo_diag_fun is not None:
            wx.CallAfter(photo_diag_fun, phDiag)
        phDiag.ShowModal()
        phDiag.Destroy()
        ind = phDiag.chosenImgInd
        if ind is None:
            # No decision, abort
            return None
        if ind == -1:
            # Add image in preview
            imgName = find_new_name(imgName, sameDateFileList, file_extension)
        else:
            # Add text to existing image
            imgName = sameDateFileList[ind]
            return os.path.join(img_folder, imgName)

    # Add file extension and copy image to project
    imgName = imgName + file_extension
    copied_file_name = os.path.join(img_folder, imgName)
    copy2(lf, copied_file_name)
    return copied_file_name


def create_dir(dir_name: str) -> None:
    """Creates the given directory recursively.
    """
    if not os.path.exists(dir_name):
        os.makedirs(dir_name)
    return


class FileDrop(wx.FileDropTarget):
    """Drop Target to drop images to be added.
    """

    origImgDate = None

    def __init__(self, window, frame):

        wx.FileDropTarget.__init__(self)
        self.window = window
        self.frame = frame
        self.loadedFile = None
        self.newFileName = None

    def OnDropFiles(self, x, y, filenames) -> bool:
        """Handle the dropped files.

        Checks if the extension is supported and stores
        the file path.
        """

        if len(filenames) > 1:
            print("Can only process one file at a time.")
        curr_file = filenames[0]
        _, ext = os.path.splitext(curr_file)
        ext = ext[1:].lower()
        if ext not in ["jpg", "jpeg", "png", "bmp"]:
            print(f"ERROR: Unsupported fucking file type: {ext}")
            return False

        # Get date and set image
        self.origImgDate = get_time_from_file(curr_file)
        self.frame.set_img_with_date(curr_file, self.origImgDate)
        self.loadedFile = curr_file

        return True


def format_date_time(date_time: wx.DateTime) -> str:
    """Format the given datetime in a string.
    """
    dt: wx.DateTime = date_time
    d, mon, y = dt.GetDay(), dt.GetMonth() + 1, dt.GetYear()
    h, m = pad_int_str(dt.GetHour()), pad_int_str(dt.GetMinute())
    s = pad_int_str(dt.GetSecond())

    str_out = f"{wx.DateTime.GetWeekDayName(dt.GetWeekDay())}, "
    str_out += f"{d}.{mon}.{y}, Time: {h}:{m}:{s}"
    return str_out


def getImageToShow(filename: str, size: int = 180, border: int = 5) -> wx.Bitmap:
    """Converts the specified image to a bitmap of according size.

    Assuming quadratic size.

    Args:
        filename: The path to the file.
        size: The size of the returned image.
        border: Border width.

    Returns:
        The bitmap with the specified size
    """
    bor_2 = 2 * border
    wxBmp = wx.Bitmap(filename)
    image = wxBmp.ConvertToImage()
    imgSize = image.GetSize()
    max_size = imgSize[0] if imgSize[0] > imgSize[1] else imgSize[1]
    fac = size / max_size
    s_diff_half = abs(imgSize[0] - imgSize[1]) * fac / 2
    image.Rescale(fac * imgSize[0], fac * imgSize[1], wx.IMAGE_QUALITY_HIGH)
    img_sz = wx.Size(size + bor_2, size + bor_2)
    image.Resize(img_sz, wx.Point(border, border + s_diff_half), 0, 0, 0)
    result = wx.Bitmap(image)
    return result


class ShowCapture(wx.Panel):
    """Panel that shows the content recorded by the webcam."""

    def __init__(self, parent, capture, fps: int = 25):

        # Capture first frame to get size
        self.capture = capture
        ret, frame = self.capture.read()
        height, width = frame.shape[:2]
        self.h_by_w = height / width
        self.win_size = (300, int(self.h_by_w * 300))

        wx.Panel.__init__(self, parent, wx.ID_ANY, (0, 0), self.win_size)
        parent.SetSize(self.win_size)

        self.h, self.w = self.win_size
        frame = cv2.cvtColor(cv2.resize(frame, self.win_size), cv2.COLOR_BGR2RGB)
        self.bmp = wx.Bitmap.FromBuffer(self.h, self.w, frame)

        self.timer = wx.Timer(self)
        self.timer.Start(1000.0 / fps)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_TIMER, self.NextFrame)

    def getCurrFrame(self):
        ret, frame = self.capture.read()
        if ret:
            return cv2.cvtColor(frame, cv2.COLOR_BGR2RGB)

    def OnPaint(self, _):
        dc = wx.BufferedPaintDC(self)
        dc.DrawBitmap(self.bmp, 0, 0)

    def NextFrame(self, _):
        ret, frame = self.capture.read()
        if ret:
            frame = cv2.cvtColor(cv2.resize(frame, self.win_size), cv2.COLOR_BGR2RGB)
            self.bmp.CopyFromBuffer(frame)
            self.Refresh()


class AcceptPhoto(TwoButtonDialogBase):
    """Dialog lets you look at the taken picture and
    decide if you want to take a new one or keep it
    """

    taken_img = None
    orig_img = None

    def __init__(self, parent=None, img=None, title="Accept photo?"):
        super().__init__(parent, title=title)

        if img is not None:
            self.set_img(img)
        self.SetSize((400, 400))
        self.SetTitle(title)
        self.accepted = False

    def InitUI(self):
        pnl = wx.Panel(self)
        self.v_box = wx.BoxSizer(wx.VERTICAL)

        # Current Image
        height, width, c = self.taken_img.shape
        self.orig_img = self.taken_img
        height, width = (300, int(height / width * 300))
        new_size = height, width
        self.taken_img = cv2.resize(self.taken_img, new_size)
        bmp = wx.Bitmap.FromBuffer(height, width, self.taken_img.tobytes())
        bmp = wx.StaticBitmap(self, -1, bmp, size=new_size)
        self.v_box.Add(bmp, proportion=0, flag=wx.ALL, border=5)

        self.setup(pnl, "Accept", "Throw")

    def set_img(self, img):
        self.taken_img = img
        self.InitUI()

    def OnTakePic(self, e):
        print("Accepting img")
        self.accepted = True
        self.Close()

    # def Cleanup(self, destroy: bool = False):
    #     self.Destroy()

    def OnClose(self, e):
        print("Cancelled Accept Dialog")
        self.Close()


class SelfieDialog(TwoButtonDialogBase):
    """Dialog that lets you take a picture with the webcam.
    """

    taken_img = None
    dt_taken = None

    _vid_capture = None
    _img_cap = None

    accept_diag: AcceptPhoto

    def __init__(self, parent=None, title="Let me take a fucking selfie."):
        super().__init__(parent, title=title)

        self.InitUI()
        self.SetSize((400, 400))
        self.SetTitle(title)

        self.accept_diag = AcceptPhoto(None, img=None)

    def InitUI(self):
        pnl = wx.Panel(self)
        self.v_box = wx.BoxSizer(wx.VERTICAL)

        # Current Image
        self._vid_capture = cv2.VideoCapture(0)
        self._img_cap = ShowCapture(self, self._vid_capture)
        self.v_box.Add(self._img_cap, proportion=0, flag=wx.ALL, border=5)
        self._img_cap.Show()

        self.setup(pnl, "Shoot", "Cancel")
        # self.Bind(wx.EVT_CLOSE, self.release_cap)

    def OnTakePic(self, e, fun: Callable = None):
        """Takes a picture with the webcam.

        Captures the current frame and
        shows the Dialog that lets the user accept the photo or decide
        to take another one.
        """
        # accept_diag = AcceptPhoto(None, img=self._img_cap.getCurrFrame())
        self.accept_diag.set_img(self._img_cap.getCurrFrame())
        if fun is not None:
            wx.CallAfter(fun)
        self.accept_diag.ShowModal()
        if self.accept_diag.accepted:
            self.taken_img = self.accept_diag.orig_img
            self.release_cap()

    def release_cap(self, _=None):
        print("Releasing capture")
        self._vid_capture.release()
        cv2.destroyAllWindows()
        self.accept_diag.Destroy()
        self.Close()

    def Destroy(self):
        self._vid_capture.release()
        cv2.destroyAllWindows()
        self.accept_diag.Destroy()
        super().Destroy()

    def OnClose(self, e):
        print("Cancelled Selfie Dialog")
        self.release_cap()



"""The main frame.

Inspired by: ZetCode wxPython tutorial, www.zetcode.com
"""

import os
import shutil
from typing import Callable, List, Union, Optional

import cv2
import wx
from pkg_resources import resource_filename

import story_time
from story_time import util
from story_time.XML_write import save_entry, find_closest_entry
from story_time.util import (
    FileDrop,
    icon_path,
    ChangeDateDialog,
    get_info_from_file,
    update_folder,
    scale_bitmap,
    format_date_time,
    getImageToShow,
    SelfieDialog,
    get_img_name_from_time,
    temp_folder,
    create_dir,
    create_xml_and_img_folder,
    write_folder_to_file,
    rep_newlines_with_space,
    copy_img_file_to_imgs,
    CustomMessageDialog,
    header_col,
    but_bg_col,
    text_bg_col,
    header_f_info,
    button_f_info,
    PhotoShow,
)

ID_MENU_PHOTO = wx.Window.NewControlId()
ID_MENU_CHANGE_DATE = wx.Window.NewControlId()
ID_MENU_SELFIE = wx.Window.NewControlId()
ID_MENU_CHOOSE_DIR = wx.Window.NewControlId()
ID_MENU_IMPORT = wx.Window.NewControlId()
ID_CLICK_BUTTON = wx.Window.NewControlId()
ID_CLICK_OK_BUTTON = wx.Window.NewControlId()
ID_CLICK_NEXT_ENTRY = wx.Window.NewControlId()
ID_CLICK_PREVIOUS_ENTRY = wx.Window.NewControlId()

LR_EXPAND = wx.LEFT | wx.RIGHT | wx.EXPAND
EXPAND_ALL = wx.ALL | wx.EXPAND


class TextLinePanel(wx.Panel):
    stat_text: wx.StaticText

    def __init__(
        self,
        parent,
        text: str = "Test Text",
        center_text: bool = False,
        bg_col: wx.Colour = "Red",
    ):
        wx.Panel.__init__(self, parent)
        box = wx.BoxSizer(wx.VERTICAL)
        s = wx.ALIGN_CENTRE_HORIZONTAL if center_text else wx.ALIGN_LEFT
        self.stat_text = wx.StaticText(self, label=text, style=s)
        self.stat_text.SetFont(wx.Font(header_f_info))
        box.Add(self.stat_text, 1, wx.ALIGN_CENTER_HORIZONTAL | EXPAND_ALL, 5)
        box.Fit(self)
        self.SetAutoLayout(True)
        self.SetSizer(box)
        self.SetBackgroundColour(bg_col)


class TwoButtonPanel(wx.Panel):
    def __init__(
        self,
        parent,
        labels: List[str],
        center: bool = True,
        bg_col: wx.Colour = "Green",
    ):
        wx.Panel.__init__(self, parent)
        box = wx.BoxSizer(wx.VERTICAL)
        # Next and previous entry buttons
        self.but_1 = wx.Button(self, wx.ID_ANY, label=labels[0], size=(70, 30))
        self.but_2 = wx.Button(self, wx.ID_ANY, label=labels[1], size=(70, 30))
        self.but_1.SetFont(wx.Font(button_f_info))
        self.but_2.SetFont(wx.Font(button_f_info))

        h_box_p_but = wx.BoxSizer(wx.HORIZONTAL)
        h_box_p_but.Add(self.but_1)
        h_box_p_but.Add(self.but_2, flag=wx.LEFT, border=5)

        flags = (
            wx.ALIGN_CENTER | wx.TOP | wx.BOTTOM
            if center
            else wx.ALIGN_RIGHT | wx.TOP | wx.BOTTOM
        )
        box.Add(h_box_p_but, 1, flags, 5)
        box.Fit(self)
        self.SetAutoLayout(True)
        self.SetSizer(box)
        self.SetBackgroundColour(bg_col)

    def set_but_methods(self, met_1, met_2):
        self.Bind(wx.EVT_BUTTON, met_1, id=self.but_1.Id)
        self.Bind(wx.EVT_BUTTON, met_2, id=self.but_2.Id)


class TextAndImgPanel(wx.Panel):
    text_box: Union[wx.TextCtrl, wx.StaticText]

    def __init__(
        self,
        parent,
        editable: bool = False,
        drop_tgt: bool = False,
        bg_col: wx.Colour = wx.Colour(200, 255, 200),
    ):
        wx.Panel.__init__(self, parent)
        box = wx.BoxSizer(wx.VERTICAL)

        # The image
        img_name = f"default_img{'_txt' if drop_tgt else ''}.png"
        # self.default_img = os.path.join(icon_path, img_name)
        self.default_img = resource_filename(__name__, f"Icons/{img_name}")

        self.bmp_shown = getImageToShow(self.default_img)
        self.img = wx.StaticBitmap(self, -1, self.bmp_shown)
        if drop_tgt:
            self.img.Hide()
            self.fileDrop = FileDrop(self.img, parent)
            self.img.SetDropTarget(self.fileDrop)
        self.img.Hide()

        # Text field
        text_shown = "Default text."
        kws = {"style": wx.TE_MULTILINE, "size": (-1, 180)}
        if editable:
            self.text_box = wx.TextCtrl(self, **kws)
        else:
            self.text_box = wx.StaticText(self, label=text_shown, **kws)
            # self.text_box = wx.TextCtrl(self, style=wx.TE_MULTILINE | wx.TE_READONLY, size=(-1, 180))
        self.text_box.SetBackgroundColour(bg_col)

        EXP_ALL = wx.EXPAND | wx.ALL
        h_box_3 = wx.BoxSizer(wx.HORIZONTAL)
        h_box_3.Add(self.text_box, proportion=3, flag=EXP_ALL, border=5)
        h_box_3.Add(self.img, proportion=1, flag=EXP_ALL, border=5)

        box.Add(h_box_3, 1, EXP_ALL, 5)
        box.Fit(self)
        self.SetAutoLayout(True)
        self.SetSizer(box)
        self.SetBackgroundColour(bg_col)


class ToolbarPanel(wx.Panel):
    tool_list = [
        ("save_icon.png", "Save", "Save entry.",),
        ("photo_icon.png", "Photo", "Change to photo mode.", True,),
        ("calendar_icon.png", "Change", "Choose another date and time.",),
        ("folder_icon.png", "Dir", "Change directory.",),
        ("webcam_icon.png", "Selfie", "Take a picture with your webcam.",),
    ]
    tools = None
    photoTool = None

    def __init__(self, parent, bg_col: wx.Colour = wx.Colour(140, 140, 255)):
        wx.Panel.__init__(self, parent)
        self.tools = []
        self.toolbar = wx.ToolBar(self, -1)
        self.setup_toolbar()
        self.SetBackgroundColour("Yellow")

        box = wx.BoxSizer(wx.VERTICAL)
        tb = self.toolbar
        box.Add(tb, 1, wx.ALIGN_RIGHT | EXPAND_ALL, 0)
        box.Fit(self)
        self.SetAutoLayout(True)
        self.SetSizer(box)
        self.toolbar.SetBackgroundColour(bg_col)

    def setup_toolbar(self) -> None:
        """Sets the toolbar up."""
        iconSize = (50, 50)

        self.toolbar.SetToolBitmapSize(iconSize)

        for ct, t in enumerate(self.tool_list):
            icon_name, name, help_txt, *rest = t
            tool_id = wx.Window.NewControlId()
            img_path = resource_filename(__name__, f"Icons/{icon_name}")
            icon = scale_bitmap(wx.Bitmap(img_path), *iconSize)
            args = (tool_id, name, icon)
            fun = (
                self.toolbar.AddCheckTool
                if rest is not None and rest
                else self.toolbar.AddTool
            )
            tool = fun(*args, shortHelp=help_txt)
            self.tools.append(tool)

        self.photoTool = self.tools[1]

        self.toolbar.AddSeparator()
        self.toolbar.Realize()

    def bind_tools(self, met_list: List[Callable]):
        assert len(met_list) == 5 == len(self.tools), f"Tools: {self.tools}"
        for met, tool in zip(met_list, self.tools):
            self.Bind(wx.EVT_TOOL, met, tool)

    pass


class StoryTimeApp(wx.Frame):
    """The Story Time App.

    This is the main frame, it contains all the functionality
    and opens other frames / dialogs when certain buttons
    are pushed.
    """

    count: int

    # GUI elements
    toolbar = None
    photoTool = None

    main_panel: Union[wx.BoxSizer, wx.Panel]
    fileDrop: FileDrop

    # Text
    input_text_field: wx.TextCtrl
    dateLabel: wx.StaticText
    cwd: wx.StaticText
    fix_text_box: wx.StaticText

    # Images
    img: wx.StaticBitmap
    image_drop_space: wx.StaticBitmap
    img_prev: wx.StaticBitmap
    prev_img_space: wx.StaticBitmap
    bmp_shown: wx.Bitmap

    prev_img_name = None

    # Boxes
    v_box: wx.BoxSizer
    h_box_1: wx.BoxSizer
    h_box_cwd: wx.BoxSizer
    h_box_4: wx.BoxSizer

    # Data to keep track of the entry in the preview. `prev_dt` contains the
    # wx.DateTime of the previewed entry. If it reaches the end, it is either
    # `max_dt` or `min_dt`.
    prev_dt: wx.DateTime
    newest_reached: Optional[bool] = None
    max_dt: wx.DateTime = wx.DateTime(1, 1, 10000)
    min_dt: wx.DateTime = wx.DateTime(1, 1, 1)

    def __init__(self, *args, **kwargs):
        super(StoryTimeApp, self).__init__(*args, **kwargs)
        self.default_img_drop = os.path.join(icon_path, "default_img_txt.png")
        self.default_img = os.path.join(icon_path, "default_img.png")
        self.cdDialog = ChangeDateDialog(None, title="Change Date of entry")
        icon = wx.Icon()
        icon.CopyFromBitmap(
            wx.Bitmap(os.path.join(icon_path, "Entwurf.jpg"), wx.BITMAP_TYPE_ANY)
        )
        self.SetIcon(icon)
        files_path = get_info_from_file()
        update_folder(files_path)
        print("util.img_folder", util.img_folder)
        self.InitUI()
        self.SetSize((700, 800))
        self.SetTitle("Story Time")
        self.Center()

        self.imgLoaded = False

    def setup_toolbar(self, panel: wx.Panel = None) -> None:
        """Sets the toolbar up."""
        iconSize = (50, 50)

        self.toolbar = self.CreateToolBar() if panel is None else panel.CreateToolBar()
        self.toolbar.SetToolBitmapSize(iconSize)

        tool_list = [
            ("save_icon.png", wx.ID_SAVE, "Save", "Save entry.", self.OnSave),
            (
                "photo_icon.png",
                ID_MENU_PHOTO,
                "Photo",
                "Change to photo mode.",
                self.OnPhoto,
                True,
            ),
            (
                "calendar_icon.png",
                ID_MENU_CHANGE_DATE,
                "Change",
                "Choose another date and time.",
                self.OnChangeDate,
            ),
            (
                "folder_icon.png",
                ID_MENU_CHOOSE_DIR,
                "Dir",
                "Change directory.",
                self.OnChangeDir,
            ),
            (
                "webcam_icon.png",
                ID_MENU_SELFIE,
                "Selfie",
                "Take a picture with your webcam.",
                self.OnSelfie,
            ),
        ]
        for ct, t in enumerate(tool_list):
            icon_name, tool_id, name, help_txt, met, *rest = t
            b_map = wx.Bitmap(os.path.join(icon_path, icon_name))
            icon = scale_bitmap(b_map, *iconSize)
            if rest is not None and rest:
                tool = self.toolbar.AddCheckTool(
                    tool_id, name, icon, shortHelp=help_txt
                )
            else:
                tool = self.toolbar.AddTool(tool_id, name, icon, shortHelp=help_txt)

            self.Bind(wx.EVT_TOOL, met, tool)
            tool_list[ct] = tool

        self.photoTool = tool_list[1]

        self.toolbar.AddSeparator()
        self.toolbar.Realize()

    def setup_one_line_static(self, lab: str, font):
        """Adds a static textbox spanning one line."""
        h_box_1 = wx.BoxSizer(wx.HORIZONTAL)
        stat_text = wx.StaticText(self.main_panel, label=lab)
        stat_text.SetFont(font)
        h_box_1.Add(stat_text, flag=wx.RIGHT, border=8)
        tc = wx.StaticText(self.main_panel)
        h_box_1.Add(tc, proportion=1)
        self.v_box.Add(h_box_1, flag=LR_EXPAND | wx.TOP, border=10)
        self.v_box.Add((-1, 10))
        return h_box_1, stat_text

    def InitUI(self) -> None:

        self.count = 5

        self.main_panel = wx.Panel(self)

        self.setup_toolbar()
        font = wx.SystemSettings.GetFont(wx.SYS_SYSTEM_FONT)
        font.SetPointSize(9)
        self.v_box = wx.BoxSizer(wx.VERTICAL)

        # Date of entry
        self.h_box_1, self.dateLabel = self.setup_one_line_static("Date", font)

        # Working directory text
        self.h_box_cwd, self.cwd = self.setup_one_line_static("This/is/a/Bug", font)

        # Text
        self.setup_one_line_static("Input text below", font)

        # Text input field
        self.input_text_field = wx.TextCtrl(
            self.main_panel, style=wx.TE_MULTILINE, size=(-1, 180)
        )

        # Drop Target
        self.bmp_shown = getImageToShow(self.default_img_drop)
        self.image_drop_space = wx.StaticBitmap(self.main_panel, -1, self.bmp_shown)
        self.image_drop_space.Hide()
        self.fileDrop = FileDrop(self.image_drop_space, self)
        self.image_drop_space.SetDropTarget(self.fileDrop)

        # Preview text field
        text_shown = "Default text."
        self.fix_text_box = wx.StaticText(
            self.main_panel, label=text_shown, style=wx.TE_MULTILINE, size=(-1, 180)
        )

        # Preview image field
        self.bmp_shown = getImageToShow(self.default_img)
        self.img_prev = wx.StaticBitmap(self.main_panel, -1, self.bmp_shown)
        self.prev_img_space = self.img_prev
        self.prev_img_space.Hide()

        # Next and previous entry buttons
        next_but = wx.Button(
            self.main_panel, ID_CLICK_NEXT_ENTRY, label="Next", size=(70, 30)
        )
        self.Bind(wx.EVT_BUTTON, self.toggle_prev_img, id=ID_CLICK_NEXT_ENTRY)
        prev_but = wx.Button(
            self.main_panel, ID_CLICK_PREVIOUS_ENTRY, label="Previous", size=(70, 30)
        )
        self.Bind(wx.EVT_BUTTON, self.toggle_prev_img, id=ID_CLICK_PREVIOUS_ENTRY)

        h_box_p_but = wx.BoxSizer(wx.HORIZONTAL)
        h_box_p_but.Add(prev_but)
        h_box_p_but.Add(next_but, flag=wx.LEFT, border=5)

        # Add stuff to horizontal boxes
        EXP_ALL = wx.EXPAND | wx.ALL
        h_box_3 = wx.BoxSizer(wx.HORIZONTAL)
        h_box_3.Add(self.fix_text_box, proportion=1, flag=EXP_ALL, border=5)
        h_box_3.Add(self.prev_img_space, proportion=0, flag=wx.ALL, border=5)
        self.h_box_4 = wx.BoxSizer(wx.HORIZONTAL)
        self.h_box_4.Add(self.input_text_field, proportion=1, flag=EXP_ALL, border=5)
        self.h_box_4.Add(self.image_drop_space, proportion=0, flag=wx.ALL, border=5)

        # Add to vertical box
        self.v_box.Add(self.h_box_4, proportion=1, flag=EXP_ALL, border=5)
        # self.v_box.Add((-1, 25))

        self.v_box.Add(h_box_p_but, flag=LR_EXPAND | wx.TOP, border=5)
        # self.v_box.Add((-1, 25))

        self.v_box.Add(h_box_3, proportion=1, flag=LR_EXPAND, border=10)
        self.v_box.Add((-1, 25))

        # Save and close buttons
        btn1 = wx.Button(
            self.main_panel, ID_CLICK_OK_BUTTON, label="Save", size=(70, 30)
        )
        self.Bind(wx.EVT_BUTTON, self.OnOKButtonClick, id=ID_CLICK_OK_BUTTON)
        btn2 = wx.Button(self.main_panel, ID_CLICK_BUTTON, label="Close", size=(70, 30))
        self.Bind(wx.EVT_BUTTON, self.OnCloseButtonClick, id=ID_CLICK_BUTTON)

        h_box_5 = wx.BoxSizer(wx.HORIZONTAL)
        h_box_5.Add(btn1)
        h_box_5.Add(btn2, flag=wx.LEFT | wx.BOTTOM, border=5)
        self.v_box.Add(h_box_5, flag=wx.ALIGN_RIGHT | wx.RIGHT, border=10)

        self.main_panel.SetSizer(self.v_box)

        self.update_preview_text()
        self.set_date_txt()
        self.set_folder_txt()

        self.Bind(wx.EVT_CLOSE, self.OnQuit)

    def set_date_txt(self):
        """Sets the text in the date textbox."""
        self.dateLabel.SetLabelText("Date: " + format_date_time(self.cdDialog.dt))

    def set_folder_txt(self):
        """Sets the text in the directory textbox."""
        self.cwd.SetLabelText("Working directory: " + story_time.util.data_path)

    def set_img(self, name: str) -> None:
        """Sets an image in the preview panel in photo mode.

        Given the path of the image.
        """
        self.bmp_shown = getImageToShow(name)
        self.image_drop_space.SetBitmap(self.bmp_shown)

    def set_prev_img(self, name: str):
        """Sets an image in the entry preview panel.

        Given the path of the image.
        """
        self.bmp_shown = getImageToShow(name)
        self.prev_img_space.SetBitmap(self.bmp_shown)
        self.prev_img_space.Show()

    def rem_prev_img(self):
        """Removes the image by hiding it."""
        if self.prev_img_space.IsShown():
            self.prev_img_name = None
            self.prev_img_space.Hide()

    def set_img_with_date(self, curr_file: str, img_date: wx.DateTime) -> None:
        """Sets an image and updates the time.
        """
        self.update_date(img_date)
        self.imgLoaded = True
        self.set_img(curr_file)

        self.resized_layout()

    def OnSave(self, *args, **kwargs) -> None:
        """Same as if the save button was clicked.
        """
        self.OnOKButtonClick(*args, **kwargs)

    def on_taken_image_clicked(self, _):
        lf = self.fileDrop.loadedFile
        if lf is not None:
            PhotoShow(self, lf).ShowModal()
            print("Enlarge taken!")
        else:
            print("No photo taken!")

    def on_prev_image_clicked(self, _):
        prev_i = self.prev_img_name
        if prev_i is not None:
            PhotoShow(self, prev_i).ShowModal()

    def OnSelfie(
        self, e, _diag_fun: Callable = None, _photo_fun: Callable = None
    ) -> None:
        """Opens dialog that shows the webcam and lets you take a picture
        with it which is added to the preview window then.
        """
        # Go to photo mode if not there yet
        if not self.photoTool.IsToggled():
            # Abort if canceled
            if self.OnPhoto(e, _photo_fun) == -1:
                self.toolbar.ToggleTool(self.photoTool.Id, False)
                return
            self.toolbar.ToggleTool(self.photoTool.Id, True)

        # Show the dialog
        sDiag = SelfieDialog()
        if _diag_fun is not None:
            wx.CallAfter(_diag_fun, sDiag)
        sDiag.ShowModal()
        sDiag.Destroy()
        img = sDiag.taken_img

        # Save the image and show it in the preview
        if img is not None:
            curr_dt = wx.DateTime.Now()
            f_name = get_img_name_from_time(curr_dt) + "_Self.png"
            f_path = os.path.join(temp_folder, f_name)
            create_dir(temp_folder)
            cv2.imwrite(f_path, cv2.cvtColor(img, cv2.COLOR_BGR2RGB))
            self.fileDrop.loadedFile = f_path
            self.set_img_with_date(f_path, curr_dt)

    @staticmethod
    def toggle_prev_img(_):
        """Toggles the preview image."""
        print("Toggling")
        pass

    def check_if_discard_changes(self, _deb_fun: Callable = None) -> int:
        """Checks whether there is unsaved info.

        Returns 0 if there is None, 1 if it may be discarded
        and -1 if the user wants to keep it."""
        textStr = self.input_text_field.GetValue()
        if textStr != "" or self.imgLoaded:
            add_string = ", the loaded image" if self.imgLoaded else ""
            msg = (
                f"If you proceed, the text{add_string} and the chosen "
                f"time will be lost. Do you want to proceed?"
            )
            md = CustomMessageDialog(
                msg,
                "Warning",
                self,
                ok_label="Fuck yeah!",
                cancel_label="No fucking way!",
            )
            if _deb_fun is not None:
                wx.CallAfter(_deb_fun, md)
            md.ShowModal()
            if not md.okay:
                return -1
            else:
                return 1
        return 0

    def OnPhoto(self, _, _deb_fun: Callable = None) -> int:
        """Change to photo mode or back.

        If there is text in the textfield or an image loaded warn
        the user that it will be lost if he continues.
        """
        print("photoTool clicked")
        discard_int = self.check_if_discard_changes(_deb_fun)
        if discard_int == -1:
            tog = self.photoTool.IsToggled()
            self.toolbar.ToggleTool(self.photoTool.Id, not tog)
            return -1
        elif discard_int == 1:
            self.remove_written_text()

        # Toggle the showing of the Image (drop space)
        if self.image_drop_space.IsShown():
            self.removeImg()
            self.imgLoaded = False
            self.image_drop_space.Hide()
        else:
            self.image_drop_space.Show()

        # Update date and time and the layout
        self.set_date_to_now()
        # self.main_panel.Layout()
        self.resized_layout()
        return 0

    def OnCloseButtonClick(self, *args, **kwargs) -> None:
        """Same as clicking X. Closes the application.
        """
        print("Close Button clicked")
        self.OnQuit(*args, **kwargs)

    def removeImg(self) -> None:
        """Set the image in the image drop space to the default.
        """
        self.fileDrop.loadedFile = None
        self.set_img(self.default_img_drop)
        self.imgLoaded = False

    def OnOKButtonClick(self, _, _no_text_fun: Callable = None) -> None:
        """Saves the text (and the photo in photo mode) in an XML entry.

        Does nothing if text (or image in photo mode) is missing.
        """

        # Check if there is any text at all
        textStr = self.input_text_field.GetValue()
        if textStr == "":
            md = CustomMessageDialog(
                "No fucking text specified!!",
                "Info",
                self,
                cancel_only=True,
                cancel_label="Got it",
            )
            if _no_text_fun is not None:
                wx.CallAfter(_no_text_fun, md)
            md.ShowModal()
            return

        # Check which mode is on
        tog = self.photoTool.IsToggled()
        if tog:
            # Check if there is an image
            lf = self.fileDrop.loadedFile
            if lf is None:
                md = CustomMessageDialog(
                    "No fucking image specified!!",
                    "Info",
                    self,
                    cancel_only=True,
                    cancel_label="Got it",
                )
                if _no_text_fun is not None:
                    wx.CallAfter(_no_text_fun, md)
                md.ShowModal()
                # wx.MessageBox("No fucking image!!", "Info", wx.OK | wx.ICON_EXCLAMATION)
                return

            # Save image entry
            curr_dat = self.cdDialog.dt
            copied_file_name = copy_img_file_to_imgs(lf, curr_dat)
            save_entry(textStr, curr_dat, "photo", copied_file_name)
        else:
            save_entry(textStr, self.cdDialog.dt)

        # Clear the contents
        self.removeImg()
        self.remove_written_text()

    def remove_written_text(self) -> None:
        """Clear text field and update date and time to now.
        """
        self.input_text_field.Clear()
        self.set_date_to_now()

    def OnChangeDate(self, _, _fun: Callable = None) -> None:
        """
        Shows dialog that lets the user change the current date.
        """
        if _fun is not None:
            wx.CallAfter(_fun, self.cdDialog)
        self.cdDialog.ShowModal()
        self.update_date()

    def OnChangeDir(self, _) -> None:
        """Shows dialog that lets the user change the current directory.
        """
        # Show dialog and get folder
        files_path = util.ask_for_dir()

        # If none selected return
        if files_path == "" or files_path is None:
            print("No new folder selected.")
            return

        # Update and create data directories if not existing
        update_folder(files_path)
        self.cwd.SetLabelText(story_time.util.data_path)
        create_xml_and_img_folder(files_path)
        self.set_date_to_now()

    def OnX(self, _, _deb_fun: Callable = None):
        """Called when the X is clicked to close.

        Also if self.Close() is called."""
        print("OnX")
        if self.check_if_discard_changes(_deb_fun) == -1:
            print("Not exiting because of unsaved data.")
            return
        self.Cleanup(None)

    def OnQuit(self, _, _deb_fun: Callable = None) -> None:
        """Closing the app, writes the working directory to file for next use,
        empties temp folder and closes the app.
        """
        print("Quit")
        if self.check_if_discard_changes(_deb_fun) != -1:
            self.Cleanup(None)

    def Cleanup(self, _) -> None:
        # Is this even used???
        print("Cleanup")
        self.cdDialog.Destroy()
        write_folder_to_file()
        if os.path.isdir(temp_folder):
            shutil.rmtree(temp_folder)
        self.Destroy()

    def set_date_to_now(self) -> None:
        """Updates the date and time to now.
        """
        self.update_date(wx.DateTime.Now())

    def update_date(self, new_date: wx.DateTime = None) -> None:
        """Updates the date to specified datetime.

        Updates the static datetime text and looks
        for the most recent XML text entries for text preview.
        """
        if new_date is not None:
            self.cdDialog.dt = new_date
        self.dateLabel.SetLabel("Date: " + format_date_time(self.cdDialog.dt))
        # The line below is probably unnecessary.
        # Actually has already been helpful in some cases!
        self.fix_text_box.SetLabel("This is a bug!")
        self.prev_dt = self.cdDialog.dt
        self.update_preview_text(set_next=False)

    def _get_text_to_put(
        self, last: bool = True, set_img: bool = False, use_prev_dt: bool = False
    ) -> str:
        search_dt = self.cdDialog.dt if not use_prev_dt else self.prev_dt
        ret_val = find_closest_entry(search_dt, not last)
        if ret_val is None:
            if use_prev_dt:
                self.newest_reached = not last
                self.prev_dt = self.min_dt if last else self.max_dt
                self.rem_prev_img()
            return ""
        self.newest_reached = None
        date, child = ret_val
        is_text = child.get("type") == "text"
        child_text = child.text if is_text else "Photo: " + child.find("text").text
        if set_img:
            if not is_text:
                prev_img_name = child.find("photo").text
                self.prev_img_name = os.path.join(
                    story_time.util.data_path, "Img", prev_img_name
                )
                self.set_prev_img(self.prev_img_name)
            else:
                self.rem_prev_img()

        self.prev_dt = date
        ret_str = "Last" if last else "Next"
        ret_str += " entry: " + format_date_time(date) + "\n\n"
        ret_str += rep_newlines_with_space(child_text) + "\n\n"
        return ret_str

    def update_preview_text(self, set_next: bool = None) -> None:
        """Fills the static datetime text with the most recent entries
        for preview.
        """

        # Construct the text to put into the preview panel.
        text_to_put = ""
        if set_next is not None:
            text_to_put += self._get_text_to_put(not set_next, True, True)
        else:
            text_to_put += self._get_text_to_put(True, True)
        if text_to_put == "":
            text_to_put = (
                f"No {'newer' if self.newest_reached else 'older'} entry present."
            )
            if set_next is None:
                self.prev_dt = self.min_dt

        # Set text and update layout
        self.fix_text_box.SetLabel(text_to_put)
        self.v_box.Layout()

    def resized_layout(self):
        self.Layout()


class StoryTimeAppUITest(StoryTimeApp):
    photoTool = None
    toolbar = None
    resized = True

    text_prev_sizer: wx.Sizer
    input_text_sizer: wx.Sizer

    def __init__(self, *args, **kwargs):
        super(StoryTimeApp, self).__init__(*args, **kwargs)
        self.default_img_drop = os.path.join(icon_path, "default_img_txt.png")
        self.default_img = os.path.join(icon_path, "default_img.png")
        self.cdDialog = ChangeDateDialog(None, title="Change Date of entry")
        icon = wx.Icon()
        icon.CopyFromBitmap(
            wx.Bitmap(os.path.join(icon_path, "Entwurf.jpg"), wx.BITMAP_TYPE_ANY)
        )
        self.SetIcon(icon)
        files_path = get_info_from_file()
        update_folder(files_path)
        print("util.img_folder", util.img_folder)
        self.InitUI()
        self.SetSize((700, 800))
        self.SetTitle("Story Time")
        self.Center()

        self.imgLoaded = False

    def next_entry(self, _):
        self.update_preview_text(True)

    def prev_entry(self, _):
        self.update_preview_text(False)

    def InitUI(self) -> None:

        # Setup toolbar
        met_list = [
            self.OnSave,
            self.OnPhoto,
            self.OnChangeDate,
            self.OnChangeDir,
            self.OnSelfie,
        ]
        tool_panel = ToolbarPanel(self, bg_col=header_col)
        tool_panel.bind_tools(met_list)
        self.photoTool = tool_panel.photoTool
        self.toolbar = tool_panel.toolbar

        path_text = TextLinePanel(
            self,
            text="Working directory: Path/to/the/working/directory",
            bg_col=header_col,
        )
        self.cwd = path_text.stat_text
        time_text = TextLinePanel(
            self,
            text="Current date and time: 12.12.1212, 23:12:55",
            center_text=True,
            bg_col=but_bg_col,
        )
        self.dateLabel = time_text.stat_text

        # Buttons
        next_prev_buttons = TwoButtonPanel(
            self, labels=["Previous", "Next"], bg_col=but_bg_col
        )

        next_prev_buttons.set_but_methods(self.prev_entry, self.next_entry)
        save_close_buttons = TwoButtonPanel(
            self, labels=["Save", "Close"], center=False, bg_col=but_bg_col
        )
        save_close_buttons.set_but_methods(
            self.OnOKButtonClick, self.OnCloseButtonClick
        )

        # Text and images
        text_edit = TextAndImgPanel(
            self, editable=True, drop_tgt=True, bg_col=text_bg_col
        )
        text_edit.SetMinSize((200, 200))
        self.fileDrop = text_edit.fileDrop
        self.image_drop_space = text_edit.img
        self.image_drop_space.Bind(wx.EVT_LEFT_DOWN, self.on_taken_image_clicked)
        self.input_text_sizer = text_edit.GetSizer()
        self.input_text_field = text_edit.text_box

        text_preview = TextAndImgPanel(self, editable=False, bg_col=text_bg_col)
        text_preview.SetMinSize((200, 200))
        self.text_prev_sizer = text_preview.GetSizer()
        self.fix_text_box = text_preview.text_box
        self.prev_img_space = text_preview.img
        self.prev_img_space.Bind(wx.EVT_LEFT_DOWN, self.on_prev_image_clicked)

        # Put it all together
        box = wx.BoxSizer(wx.VERTICAL)
        box.Add(path_text, 0, LR_EXPAND)
        box.Add(tool_panel, 0, LR_EXPAND)
        box.Add(time_text, 0, LR_EXPAND)
        box.Add(text_edit, 1, LR_EXPAND)
        box.Add(save_close_buttons, 0, LR_EXPAND)
        box.Add(text_preview, 1, LR_EXPAND)
        box.Add(next_prev_buttons, 0, LR_EXPAND)
        box.Fit(self)
        self.v_box = box
        self.main_panel = box
        self.SetSizer(box)

        # Prepare UI
        self.update_preview_text()
        self.set_date_txt()
        self.set_folder_txt()
        self.Bind(wx.EVT_CLOSE, self.OnX)

        self.Bind(wx.EVT_SIZE, self.on_resize)
        self.Bind(wx.EVT_IDLE, self.OnIdle)

        self.SetMinSize((400, 600))

    def on_resize(self, _):
        self.resized = True  # set dirty
        self.Layout()

    def OnIdle(self, _):
        if self.resized:
            lf = self.fileDrop.loadedFile
            s_min = min(self.input_text_sizer.Size)
            if s_min > 30:
                set_img = self.default_img_drop if lf is None else lf
                self.image_drop_space.SetBitmap(getImageToShow(set_img, s_min - 30))

            if self.prev_img_name is not None:
                s_min = min(self.text_prev_sizer.Size)
                if s_min > 30:
                    self.prev_img_space.SetBitmap(
                        getImageToShow(self.prev_img_name, s_min - 30)
                    )

            self.Layout()
            self.resized = False  # reset the flag

    def resized_layout(self):
        self.on_resize(None)
        self.OnIdle(None)

    def set_prev_img(self, name: str):
        """Sets an image in the entry preview panel.

        Given the path of the image.
        """
        super().set_prev_img(name)

        self.resized_layout()

    pass



import setuptools

with open("README.md") as f:
    long_description = f.read()

with open("version.txt", "r", encoding="utf8") as f:
    v_num = f.read().strip()

setuptools.setup(
    name="story-time",
    version=v_num,
    description="Test",
    packages=setuptools.find_packages(exclude=["docs", "tests"]),
    include_package_data=True,
    package_data={"story_time": ["*.png", "*.txt"],},
    url="https://github.com/chbauman/StoryTime",
    author="Christian Baumann",
    author_email="chris.python.notifyer@gmail.com",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=">=3.6",
    install_requires=["wxPython", "opencv-python", "numpy", "Pillow", "six",],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Operating System :: OS Independent",
        "Topic :: Utilities",
    ],
    entry_points={"console_scripts": ["story-time=story_time.main:main"],},
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This module contains the XML reading and writing utilities.

All the text is stored in an XML file and the images are referenced
in the file.
"""
import os
import xml.etree.cElementTree as elTree
from typing import Tuple, Optional

import wx

from story_time import util


def init_XML(comm: str, year: int) -> elTree.ElementTree:
    """Initializes an XML document for a new year.

    Args:
        comm: A comment for the beginning of the year.
        year: The year of the document.

    Returns:
        XML element tree.
    """
    root = elTree.Element("root")
    head = elTree.SubElement(root, "head")
    elTree.SubElement(head, "year").text = str(year)
    elTree.SubElement(head, "text").text = comm
    elTree.SubElement(root, "doc")
    return elTree.ElementTree(root)


def insert_text_entry(doc: elTree.Element, date_time: wx.DateTime, text: str) -> None:
    """Inserts a text entry into the XML element tree.

    Args:
        doc: The doc element. Entry will be inserted as child.
        date_time: The DateTime of the entry.
        text: The text of the entry.
    """
    dt = date_time.FormatISOCombined()
    elTree.SubElement(doc, "entry", date_time=dt, type="text").text = text


def insert_photo_entry(
    doc: elTree.Element, date_time: wx.DateTime, img_filename: str, text: str
) -> None:
    """Inserts a photo entry element as a child of the elTree doc.

    Args:
        doc: The doc element. Entry will be inserted as child.
        date_time: The DateTime of the entry.
        img_filename: The file name referring to the image.
        text: The text of the entry.
    """
    # TODO: Check if file exists?
    ent = elTree.SubElement(
        doc, "entry", date_time=date_time.FormatISOCombined(), type="photo"
    )
    elTree.SubElement(ent, "text").text = text
    elTree.SubElement(ent, "photo").text = img_filename


def load_XML(year: int, create: bool = True) -> Tuple[elTree.ElementTree, str]:
    """Loads an existing XML file and returns the Element tree.

    If it does not exist, a new element tree is initializes.

    Args:
        year: The year specifying the XML file.
        create: Whether to return a new tree if file not found.

    Returns:
        The element tree and the filename.
    """
    yearStr = str(year)
    xml_file = os.path.join(util.xml_folder, f"{yearStr}.xml")
    if os.path.exists(xml_file):
        tree = elTree.parse(xml_file)
    elif create:
        tree = init_XML("Das isch es Johr " + yearStr + ".", year)
    else:
        raise FileNotFoundError("File not found and create == False")
    return tree, xml_file


def save_entry(
    comm: str,
    date_time: wx.DateTime,
    entry_type: str = "text",
    img_filename: str = None,
) -> None:
    """Reads the XML file and adds an entry element with the specified content.

    Then saves the tree back to the file.

    Args:
        comm: The text for the entry.
        date_time: The DateTime of the entry.
        entry_type: The type of the entry, either "text" or "photo".
        img_filename:
    """
    year = date_time.GetYear()
    tree, xml_file = load_XML(year)

    doc = tree.getroot().find("doc")
    if entry_type == "text":
        insert_text_entry(doc, date_time, comm)
    elif entry_type == "photo":
        insert_photo_entry(doc, date_time, os.path.basename(img_filename), comm)
    else:
        raise ValueError(f"Entry of type: {entry_type} is not supported!")
    tree.write(xml_file)


def find_next_xml_file(
    year: int, newer: bool = False
) -> Optional[Tuple[int, elTree.ElementTree]]:
    """Finds the most recent XML file before year 'year'.

    If there is none, returns None, otherwise, the found year and the tree of the XML
    is returned.
    If `newer` == True, then it finds the next newer one

    Args:
        year:
        newer:

    Returns:
        The year of the closest entry and the element tree of the
        closes XML file.
    """
    closest_year = 100000 if newer else -100000
    for f in os.listdir(util.xml_folder):
        y = int(f.split(".")[0])
        if not newer and year > y > closest_year:
            closest_year = y
        if newer and year < y < closest_year:
            closest_year = y
    if abs(closest_year) == 100000:
        return None
    return closest_year, load_XML(closest_year, False)[0]


def find_closest_entry_in_tree(
    tree: elTree.ElementTree, date_time: wx.DateTime, newer: bool = False
) -> Optional[Tuple[wx.DateTime, elTree.Element]]:
    """Finds entry closest to `date_time` in doc.

    If there is no earlier entry than 'date_time' returns None.
    If newer == True, then it finds the next newer one

    Args:
        tree: The element tree to search.
        date_time: The search date and time.
        newer: Whether to find the closest newer entry.

    Returns:
        The date of the found entry and the entry element.
    """
    doc = tree.getroot().find("doc")
    temp = wx.DateTime(1, 1, 100000 if newer else 0)
    curr_child = None
    for child in doc:
        wxDT = wx.DateTime()
        wxDT.ParseISOCombined(child.get("date_time"))
        if (not newer and temp < wxDT < date_time) or (
            newer and temp > wxDT > date_time
        ):
            temp = wxDT
            curr_child = child
    if abs(temp.GetYear() - 50000) == 50000:
        return None
    return temp, curr_child


def find_closest_entry(
    dt: wx.DateTime, newer: bool = False
) -> Optional[Tuple[wx.DateTime, elTree.Element]]:
    """Finds the latest entry before 'dt', if any.

    Searches all available XML files.

    Args:
        dt: The search date and time.
        newer: Whether to find the closest newer entry.

    Returns:
        The date of the found entry and the entry element.
        If there is no next entry, None is returned.
    """
    curr_year = dt.GetYear()

    # Load XML file with specified year
    tree = None
    try:
        tree = load_XML(curr_year, False)
    except FileNotFoundError:
        pass
    date_child = None
    if tree is not None:
        tree = tree[0]
        date_child = find_closest_entry_in_tree(tree, dt, newer)
        if date_child is None:
            tree = None

    # Entry not found yet, search next XML files
    if tree is None:
        date_child = None
        while date_child is None:
            res = find_next_xml_file(curr_year, newer)
            if res is None:
                # No next file found, return None
                return None
            tree = res[1]
            curr_year = res[0]
            date_child = find_closest_entry_in_tree(tree, dt, newer)

    # Return
    assert date_child is not None, f"Terrible bug is happening!"
    date, child = date_child
    return date, child



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""The main script with the StoryTimeApp.

Run this file to run the app.

TODO:
    - PDF generation!
    - Fix webcam glitches.
    - Resize images (and text?) when resizing window.
    - Indicate diary entries in calendar.
    - Smoother UI changes (e.g. image loading)
    - Host documentation somewhere.
    - Adjust colors.
    - Achieve 100% test coverage.
    - Preview: Resolve issue with multiple entries on same Date and time.
    - Center toolbar?
    - Better fonts (size and family!)
    - Color buttons!
    - Make header color changeable: https://stackoverflow.com/questions/51000320/wxpython-change-the-headers-color
    - Handle image deletion while app is running
    - Handle app resizing! (Similar as for the photo enlargement)
    - Remove prints for non-debug scenario
    - En-/ Disable Next/Previous button if no other entry present
    - Handle non-square images in large view
    - Change black image border to better color
"""

import wx

from story_time.user_interface import StoryTimeAppUITest


def main():
    """DO ALL THE STUFF.
    """
    app = wx.App()
    ex = StoryTimeAppUITest(None)
    ex.Show()
    app.MainLoop()


if __name__ == "__main__":
    main()
